/*
  LockExtensions - Plugin for KeePass 1
  Copyright (C) 2010 Nikolaus Hammler <niki@hammler.info>

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/

#include "LockExtensions.h"
#include <tchar.h>
#include <assert.h>

static IKpAPI *g_pAPI = NULL;
static HINSTANCE g_hDLL = NULL;
static HWND g_hWnd = NULL;
static long g_iTimeout = 0;
static BOOL g_bLockOnStandby = FALSE;
static WORD g_wHotKey = 0;
static int g_iHotKeyID = 0;

KP_IMPL_STDREFIMPL(LockExtensions)

BOOL WINAPI DllMain(HINSTANCE hinstDLL, DWORD fdwReason, LPVOID lpvReserved)
{
	g_hDLL = hinstDLL;
	UNREFERENCED_PARAMETER(fdwReason);
	UNREFERENCED_PARAMETER(lpvReserved);
	return TRUE;
}

KP_EXPORT HRESULT KP_API KP_I_INITIALIZELIB_DECL(IKpUnknown* pAPI)
{
	g_iHotKeyID = GlobalAddAtom(_T("327BE89E-252E-420d-8DE2-BDF43D3DA5C8"));
	UNREFERENCED_PARAMETER(pAPI);
	return S_OK;
}

KP_EXPORT HRESULT KP_API KP_I_RELEASELIB_DECL(IKpUnknown* pAPI)
{
	DestroyWindow(g_hWnd);
	GlobalDeleteAtom(g_iHotKeyID);

	UNREFERENCED_PARAMETER(pAPI);
	SAFE_RELEASE(g_pAPI);
	return S_OK;
}

KP_EXPORT HRESULT KP_API KP_I_CREATEINSTANCE_DECL(REFIID riid, void** ppvObject, IKpUnknown* pAPI)
{
	*ppvObject = NULL;

	if(pAPI == NULL) return E_UNEXPECTED;

	if((riid == IID_IKpUnknown) || (riid == IID_IKpPlugin))
	{
		if(pAPI->QueryInterface(IID_IKpAPI, (void**)&g_pAPI) != S_OK)
			return E_FAIL;

		if(!StartWindow())
			return E_FAIL;

		IKpPlugin* pImpl = new LockExtensions();
		*ppvObject = ((riid == IID_IKpPlugin) ? pImpl : (IKpUnknown*)pImpl);
		return S_OK;
	}

	return E_NOINTERFACE;
}

void LockKeepass()
{
	if(g_pAPI->IsFileOpen())
	{
		HWND hKPWnd = g_pAPI->GetMainWindowHandle();
		assert( hKPWnd );
		SendMessage(hKPWnd, WM_COMMAND, ID_FILE_LOCK, 0);
	}
}

BOOL RegisterGlobalHotkey(WORD wHotKey)
{
	UINT uModifiers = 0;
	if((HIBYTE(wHotKey) & HOTKEYF_ALT)) uModifiers |= MOD_ALT;
	if((HIBYTE(wHotKey) & HOTKEYF_SHIFT)) uModifiers |= MOD_SHIFT;
	if((HIBYTE(wHotKey) & HOTKEYF_CONTROL)) uModifiers |= MOD_CONTROL;

	return RegisterHotKey(g_hWnd, g_iHotKeyID, uModifiers, LOBYTE(wHotKey));
}

LRESULT CALLBACK WindowProc(HWND hwnd, UINT msg, WPARAM wParam, LPARAM lParam)
{
	static DWORD lastInput = 0;
	static int activeCount = 0;

    switch(msg)
    {
		case WM_TIMER:
			{
				LASTINPUTINFO plii;
				plii.cbSize = sizeof(LASTINPUTINFO);
				plii.dwTime = 0;
				GetLastInputInfo(&plii);

				if(plii.dwTime == lastInput)
					activeCount++;
				else
					activeCount = 0;

				lastInput = plii.dwTime;

				//if(activeCount >= ((g_iTimeout<=1) ? 10 : g_iTimeout))
				if(activeCount >= g_iTimeout)
				{
					activeCount = 0;
					LockKeepass();
				}
				return TRUE;
			}
			break;
		case WM_HOTKEY:
			LockKeepass();
			return TRUE;
		case WM_POWERBROADCAST:
			if(wParam == PBT_APMSUSPEND && g_bLockOnStandby)
			{
				LockKeepass();
				return TRUE;
			}
			break;
		case WM_CLOSE:
			break;
		default:
			return DefWindowProc(hwnd, msg, wParam, lParam);
	}
	return 0;
}

BOOL StartWindow()
{
	WNDCLASSEX wc;

	wc.cbSize        = sizeof(WNDCLASSEX);
	wc.style         = CS_CLASSDC;
	wc.lpfnWndProc   = WindowProc;
	wc.cbClsExtra    = 0;
	wc.cbWndExtra    = 0;
	wc.hInstance     = GetModuleHandle(NULL);
	wc.hIcon         = 0;
	wc.hCursor       = 0;
	wc.hbrBackground = 0;
	wc.lpszMenuName  = 0;
	wc.lpszClassName = WND_CLASSNAME;
	wc.hIconSm       = 0;

	if(!RegisterClassEx(&wc))
		return FALSE;
	
	g_hWnd = CreateWindowEx(0, WND_CLASSNAME, _T(""), 0, 0, 0, 0, 0, NULL, NULL, GetModuleHandle(NULL), NULL);
	
	if(!g_hWnd)
		return FALSE;

	return TRUE;
}

LockExtensions::LockExtensions()
{
	KP_IMPL_CONSTRUCT;
}

STDMETHODIMP LockExtensions::QueryInterface(REFIID riid, void** ppvObject)
{
	KP_SUPPORT_INTERFACE(IID_IKpUnknown, IKpUnknown);
	KP_SUPPORT_INTERFACE(IID_IKpPlugin, IKpPlugin);

	*ppvObject = NULL;
	return E_NOINTERFACE;
}

void LockExtensions::LoadSettings()
{
	g_iTimeout = DEFAULT_TIMEOUT;

	IKpConfig *pCfg = NULL;
	if(g_pAPI->CreateInstance(CLSID_KpConfig, IID_IKpConfig, (void**)&pCfg) != S_OK)
		return;
	
	assert( pCfg );
	char str_cfg[SI_REGSIZE];
	
	if(pCfg->Get(CONFIG_ITEM_TIME, str_cfg))
		g_iTimeout = atol(str_cfg);

	if(pCfg->Get(CONFIG_ITEM_HOTKEY, str_cfg))
		g_wHotKey = (WORD)atol(str_cfg);

	g_bLockOnStandby = pCfg->GetBool(CONFIG_ITEM_STANDBY, DEFAULT_STANDBY);

	pCfg->Release();
}

void LockExtensions::SaveSettings()
{
	IKpConfig *pCfg = NULL;
	if(g_pAPI->CreateInstance(CLSID_KpConfig, IID_IKpConfig, (void**)&pCfg) != S_OK)
		return;
	
	assert( pCfg );

	char str_cfg[SI_REGSIZE];
	// FIXXME: do it better
	wsprintf(str_cfg, "%d", g_iTimeout);
	pCfg->Set(CONFIG_ITEM_TIME, str_cfg);

	wsprintf(str_cfg, "%lu", g_wHotKey);
	pCfg->Set(CONFIG_ITEM_HOTKEY, str_cfg);
	pCfg->SetBool(CONFIG_ITEM_STANDBY, g_bLockOnStandby);
	pCfg->Release();
}

INT_PTR CALLBACK SettingsWindowProc(HWND hwndDlg, UINT uMsg, WPARAM wParam, LPARAM lParam)
{
	switch(uMsg)
	{
		case WM_INITDIALOG:
			SetDlgItemInt(hwndDlg, IDC_IDLETIME, g_iTimeout, FALSE);
			SetFocus(GetDlgItem(hwndDlg, IDC_IDLETIME));
			CheckDlgButton(hwndDlg, IDC_LOCK_STANDBY, g_bLockOnStandby ? BST_CHECKED : BST_UNCHECKED);

			SendMessage(GetDlgItem(hwndDlg, IDC_HOTKEY), HKM_SETRULES, HKCOMB_NONE | HKCOMB_S, HOTKEYF_CONTROL | HOTKEYF_ALT);
			if(g_wHotKey)
			{
				UnregisterHotKey(g_hWnd, g_iHotKeyID);
				SendMessage(GetDlgItem(hwndDlg, IDC_HOTKEY), HKM_SETHOTKEY, g_wHotKey, 0);
			}
			return TRUE;
		case WM_COMMAND:
			if(wParam == IDOK)
			{
				BOOL bTranslated;
				long timeout = GetDlgItemInt(hwndDlg, IDC_IDLETIME, &bTranslated, FALSE);
				if(bTranslated)
					g_iTimeout = timeout;

				switch(IsDlgButtonChecked(hwndDlg, IDC_LOCK_STANDBY)) {
					case BST_CHECKED: g_bLockOnStandby = TRUE; break;
					case BST_UNCHECKED: g_bLockOnStandby = FALSE; break;
				}

				g_wHotKey = (WORD)SendMessage(GetDlgItem(hwndDlg, IDC_HOTKEY), HKM_GETHOTKEY, 0, 0);

				if(g_wHotKey)
				{
					if(!RegisterGlobalHotkey(g_wHotKey))
					{
						MessageBox(hwndDlg, _T("Could not set the hot key. Please choose a different"), _T("Unable to use hot key"), MB_ICONERROR | MB_OK);
						return FALSE;
					}
				}
				else
				{
					UnregisterHotKey(g_hWnd, g_iHotKeyID);
				}
				EndDialog(hwndDlg, 0);
			}
			else if(wParam == IDCANCEL)
			{
				EndDialog(hwndDlg, 0);
			}

			return TRUE;
	}
	return FALSE;
}

BOOL CALLBACK FindPluginWindow(HWND hwnd, LPARAM lParam)
{
	char title[1024];
	GetWindowText(hwnd, title, 1023);
	if(!strcmp(title, "Plugins"))
	{
		HWND *hDst = (HWND*)lParam;
		*hDst = hwnd;
		return FALSE;
	}
	return TRUE;
}

STDMETHODIMP_(BOOL) LockExtensions::OnMessage(DWORD dwCode, LPARAM lParamW, LPARAM lParamL)
{
	UNREFERENCED_PARAMETER(lParamW);
	UNREFERENCED_PARAMETER(lParamL);

	switch(dwCode)
	{
		case KPM_DELAYED_INIT:
			LoadSettings();
			if(!RegisterGlobalHotkey(g_wHotKey))
				MessageBox(NULL, _T("The hot key could not be registered. Please select a new one in the settings"), PLUGIN_NAME, MB_OK | MB_ICONWARNING);
			break;
		case KPM_DIRECT_CONFIG:
			{
				HWND hWnd = g_pAPI->GetMainWindowHandle();
				EnumThreadWindows(GetCurrentThreadId(), FindPluginWindow, (LPARAM)&hWnd);
				LoadSettings();
				DialogBox(g_hDLL, MAKEINTRESOURCE(IDD_SETTINGS), hWnd, SettingsWindowProc);
				SaveSettings();
			}
			break;
		case KPM_PLUGIN_INFO:
			MessageBox(NULL, _T("(c) 2010 Nikolaus Hammler\n\nhttp://niki.hammler.info"), PLUGIN_NAME, MB_OK | MB_ICONINFORMATION);
			break;
		case KPM_OPENDB_POST:
			LoadSettings();
			SetTimer(g_hWnd, TIMER_ID, 60000, NULL);
			break;
		case KPM_FILE_CLOSE_POST:
		case KPM_FILE_EXIT_PRE:
			KillTimer(g_hWnd, TIMER_ID);
			break;
		default:
			break;
	}

	return TRUE;
}

STDMETHODIMP_(LPCTSTR) LockExtensions::GetProperty(LPCTSTR lpName)
{
	UNREFERENCED_PARAMETER(lpName);
	return NULL;
}

STDMETHODIMP LockExtensions::SetProperty(LPCTSTR lpName, LPCTSTR lpValue)
{
	UNREFERENCED_PARAMETER(lpName);
	UNREFERENCED_PARAMETER(lpValue);
	return E_FAIL;
}

STDMETHODIMP_(DWORD) LockExtensions::GetMenuItemCount()
{
	return 0;
}

STDMETHODIMP_(KP_MENU_ITEM*) LockExtensions::GetMenuItems()
{
	return NULL;
}

